import pygame
import math
import random

#Увы, так и не смог найти способ адекватного подсчёта отскоков
#Либо "дребезг", либо оно неограниченно растёт, либо всё и сразу

SCREEN_SIZE = S_WIDTH, S_HEIGHT = 1280, 720 #Точки
WORLD_SIZE = W_WIDTH, W_HEIGHT = 32000, 18000 #Километры

EPS = 0.01 #Нелья просто так взять и сравнить вещественные числа
GRAVITY = 6.67 * 10**(-7) #Гравитационная постоянная
N = 10 #Число шариков


def scale(v, new_min, new_max, old_min, old_max):
    """Преобразование координаты в соответствии с заданными параметрами"""
    return new_max * (v-old_min)/(old_max-old_min)

def worldToScreen(x, y):
    """Преобразование мировых координат в экранные"""
    new_x = int(scale(x, 0, S_WIDTH, -W_WIDTH//2, W_WIDTH//2)) 
    new_y = int(scale(y, 0, S_HEIGHT, -W_HEIGHT//2, W_HEIGHT//2)) 
    return [new_x, new_y]

def screenToWorld(x, y):
    """Преобразование экранных координат в мировые"""
    new_x = (x/S_WIDTH) * W_WIDTH - W_WIDTH//2
    new_y = (y/S_HEIGHT) * W_HEIGHT - W_HEIGHT//2
    return [new_x, new_y]
    
def resize(surface, rad):
    """Масштабирование спрайта под размеры объекта"""
    scale_coeff = S_WIDTH/W_WIDTH
    new_size = [int(rad*scale_coeff)*2]*2
    return pygame.transform.smoothscale(surface, new_size)
    pass


class Object:
    """В игре два типа объектов, различаются они только по одному свойству"""
    def __init__(self, pos, rad, sprite):
        self.pos = pos
        self.rad = rad
        self.sprite = resize(sprite, rad)
        
    def draw(self, surface):
        offset = self.sprite.get_size()
        coords = worldToScreen(*self.pos)
        surface.blit(self.sprite, (coords[0] - offset[0]//2, coords[1] - offset[1]//2))
        
class Ball(Object):
    """Шарик, его можно таскать по экрану и бросать"""
    def __init__(self, pos, rad, sprite, speed):
        Object.__init__(self, pos, rad, sprite)
        self.speed = speed
        self.acceleration = pygame.math.Vector2(0, 0)
        self.drag = False

    def move(self):
        self.pos += self.speed

    def collideWalls(self):
        #Заменить pos и speed на pygame.math.vector2
        friction = 0.9
        if self.pos[0] >= W_WIDTH//2 - self.rad or self.pos[0] <= -W_WIDTH//2 + self.rad:
            self.speed[0] *= -1
            self.speed *= friction
            
        if self.pos[1] >= W_HEIGHT//2 - self.rad or self.pos[1] <= -W_HEIGHT//2 + self.rad:
            self.speed[1] *= -1
            self.speed *= friction


    def collide(self, another):
        """Проверка столкновений для шариков и планет"""
        #Вектор сонаправлен нормальной составляющей скорости первого шарика
        #длина - расстояние между центрами
        dist = another.pos - self.pos
        if self.rad + another.rad >= pygame.math.Vector2.length(dist):
            if type(another) == Ball:
                #Начинается мясо
                #Нельзя просто так взять и посчитать новую нормальную составляющую
                #Скорости при различных комбинациях почему-то всё время увеличиваются
                #При попытке собрать вектор из нормальной и тангенциальной составляющей
                #игра ведёт себя стабильно, и это хорошо
                sAngle = self.speed.angle_to(dist)
                aAngle = another.speed.angle_to(dist * (-1))

                selfNormal = pygame.math.Vector2.normalize(another.pos - self.pos)
                anotherNormal = pygame.math.Vector2.normalize(self.pos - another.pos)

                selfNormalLength = int(self.speed.length() * math.cos(sAngle))
                anotherNormalLength = int(another.speed.length() * math.cos(aAngle))
                newNormal = abs(selfNormalLength - anotherNormalLength)

                selfTangent = selfNormal.rotate(-90)
                anotherTangent = anotherNormal.rotate(-90)

                selfTangentLength = math.sqrt(self.speed.length_squared() - selfNormalLength**2)
                anotherTangentLength = math.sqrt(another.speed.length_squared() - anotherNormalLength**2)
                
                self.speed = anotherNormal * newNormal + selfTangent * selfTangentLength
                another.speed = selfNormal * newNormal + anotherTangent * anotherTangentLength
                
            elif type(another) == Planet:
                #Вычисляем новое направление нормальной составляющей скорости
                normal = pygame.math.Vector2.normalize(dist)
                #Практически магия. А вот если мы сталкиваемся с шариком, то придётся повозиться
                self.speed.reflect_ip(normal)
                #Добавить зависимость затухания от угла
                self.speed  *= another.friction
class Planet(Object):
    """Планета притягивает шарики к себе"""
    def __init__(self, pos, rad, mass, sprite, friction):
        Object.__init__(self, pos, rad, sprite)
        self.mass = mass
        self.friction = friction


def genBall(imagelist):
    rad = random.randint(400, 600)
    pos = pygame.math.Vector2(random.randint(-5000, 4000), random.randint(-6000, 6000))
    speed = pygame.math.Vector2(random.sample(range(10, 30), 2))
    image = random.choice(imagelist)
    return Ball(pos, rad, image, speed)

pygame.init()
screen = pygame.display.set_mode(SCREEN_SIZE)

#Игра обновляет своё состояние TICKS_PER_SECOND раз в секунду
TICKS_PER_SECOND = 60
#Промежуток между двумя обновлениями
STEP = 1000//TICKS_PER_SECOND

#Попытка реализовать независимые друг от друга обновление и отрисовку с треском провалилась
#В бесконечном цикле нахдился вложенный, обновлявший состояние игры TICKS_PER_SECOND раз в секунду
#После начиналась отрисовка. При таком подходе игра "висла" намертво
#Посему возвращаемся к проверенной временем реализации, где FPS и скорость обновления намертво связаны
pygame.time.set_timer(pygame.USEREVENT+1, STEP)

IS_RUNNING = True

filelist = [open("Planet0" + str(i) + ".png", "rb") for i in range(6)]
imagelist = [pygame.image.load(i).convert_alpha() for i in filelist]

balls = [genBall(imagelist) for i in range(N)]
planets = [Planet(pygame.math.Vector2(-8000, 0), 2000, 1.3 * 10**13, imagelist[3], 0.8), Planet(pygame.math.Vector2(8000, 0), 3000, 2 * 10**13, imagelist[2], 0.8)]


#Мы не можем тащить шарик бесконечно долго
F_TIME = 0
#Какую долю секунды мы тащим шарик
PART = 5


while IS_RUNNING:
    e = pygame.event.poll()
    if e.type == pygame.QUIT:
        IS_RUNNING = False
    elif e.type == pygame.MOUSEBUTTONDOWN and e.button == 1:
        for b in balls:
            if (b.pos - screenToWorld(*e.pos)).length() <= b.rad:
                F_TIME = pygame.time.get_ticks()
                b.drag = True
    elif e.type == pygame.MOUSEBUTTONUP and e.button == 1:
        for b in balls:
            b.drag = False
    elif e.type == pygame.MOUSEMOTION and e.buttons[0]:
        for b in balls:
            if b.drag:
                b.pos = pygame.math.Vector2(screenToWorld(*e.pos))
                new_speed = [i*5 for i in e.rel]
                b.speed = pygame.math.Vector2(*new_speed)
            if pygame.time.get_ticks() - F_TIME > 1000//PART:
                b.drag = False
                F_TIME = 0
                pygame.event.post(pygame.event.Event(pygame.MOUSEBUTTONUP, button=1))
    elif e.type == pygame.USEREVENT + 1:
        #Е-е-е-е, физика!
        for i in range(len(balls) - 1):
            balls[i].move()
            balls[i].collideWalls()
            for j in range(i+1, len(balls)):
                balls[i].collide(balls[j])

        for b in balls:
            b.move()
            b.collideWalls()
            for p in planets:
                b.collide(p)
                dist = p.pos - b.pos
                if dist.length() >= b.rad + p.rad:
                    length = (GRAVITY * p.mass)/pygame.math.Vector2.length_squared(dist)
                    acceleration = pygame.math.Vector2.normalize(dist)*length
                    b.speed += acceleration
    else:
        screen.fill(pygame.Color("black"))
        for b in balls: 
            b.draw(screen)
        for p in planets:
            p.draw(screen)
        pygame.display.flip()
